package jgui.ui;

import waba.ui.*;
import waba.fx.*;
import waba.sys.*;
import waba.util.*;
import extra.ui.*;

import jgui.util.*;

/**
 * The <code>TextArea</code> is a non-editable, scrollable
 * area of text that can be added to a Scroller.
 * It wraps the data provided attempting to display the text
 * without scrollbars but if there is no whitespace to wrap a
 * line at, the horizontal scrollbar must be displayed.
 *
 * @version $Header: $
 * @author Jim Guistwite
 */

// Ken Turner <kjt@cs.stir.ac.uk>: bug fixes, rewrote "setText" and added
// "setFormattedText" for efficiency

public class TextArea extends Control implements Scrollable
{
  private Dimension pSize;
  private Vector data;
  private Font font;
  private Scroller scroller;

  private int verticalScrollInc;
  private int horizontalScrollInc;
  private int preferredHeight;
  private int preferredWidth;

  private int vOffset = 0;
  private int hOffset = 0;

  /**
   * Create a new text area with no text.
   */
  public TextArea() {
    this(null);
  }

  /**
   * Create a new text area with the text provided.
   * This is the same as calling the empty constructor
   * then calling setText().
   */
  public TextArea(String text) {
    data = new Vector();
    font = new Font("Helvetica", Font.PLAIN, 12);

    // compute the horizontal and vertical scroll increments
    FontMetrics fm = getFontMetrics(font);
    verticalScrollInc = fm.getHeight();
    horizontalScrollInc = (fm.getTextWidth("Wm") / 2);

    if (text != null)
      setText(text);
  }

  /**
   * Set the font to be used to display the text.
   * The programmer should call this before calling
   * set text or should call setText again after in
   * order to recalculate the text wrapping.
   *
   * @param font font to be used to display the text.
   */
  public void setFont(Font font) {
    super.setFont(font);
    this.font = font;

    FontMetrics fm = getFontMetrics(font);
    verticalScrollInc = fm.getHeight();
    horizontalScrollInc = (fm.getTextWidth("Wm") / 2);

    Control parent = getParent();
    if ((parent != null) && (parent instanceof Scroller)) {
      //System.out.println("scroller invalidate");
      ((Scroller)parent).invalidate();
    }
  }

  // KJT 07/08/01: rewrote setText for efficiency and to handle "\n\n"

  /**
   * Set the text that should be displayed in this text area.
   * The text is wrapped at newlines and spaces.
   *
   * @param s text to be displayed.
   */
  public void setText(String s) {
    data = new Vector();

    int maxWidth = 0;

    // break sb into correctly sized chunks.
    FontMetrics fm = getFontMetrics(font);

    StringTokenizer tok = new StringTokenizer(s, " \n");
    String soFar = "";
    String next;
    String nextSpace;
    String textWithNext;
    int widthWithNext;
    int priorWidth = 0;

    while(tok.hasMoreTokens()) {
      next = tok.nextToken();
      if (priorWidth == 0) {
        widthWithNext = fm.getTextWidth(next);
	textWithNext = next;
      }
      else {
        nextSpace = " " + next;
        widthWithNext = priorWidth + fm.getTextWidth(nextSpace);
	textWithNext = soFar + nextSpace;
      }

      if (widthWithNext > width) {
	if (priorWidth == 0) {
	  // do it anyway
	  if (widthWithNext > maxWidth) {
	    maxWidth = widthWithNext;
	    //System.out.println("max width became " + maxWidth + " at 1");
	  }

	  data.add(next);
	  priorWidth = 0;
	  soFar = "";
	}
	else {
	  if (priorWidth > maxWidth) {
	    maxWidth = priorWidth;
	    //System.out.println("max width became " + maxWidth + " at 2");
	  }
	  data.add(soFar);
          if (tok.getLastDelimeter() != '\n') {		// excess ends normally?
	    priorWidth = fm.getTextWidth(next);		// start new line
	    soFar = next;
          }
          else {					// excess ends with NL
	    data.add(next);				// add what there is
	    priorWidth = 0;
	    soFar = "";
          }
	}
      }
      else {
	if (tok.getLastDelimeter() == '\n') {
	  // break here!
	  if (widthWithNext > maxWidth) {
	    maxWidth = widthWithNext;
	    //System.out.println("max width became " + maxWidth + " at 2");
	  }
	  data.add(textWithNext);
	  priorWidth = 0;
	  soFar = "";
	}
	else {
	  priorWidth = widthWithNext;
	  soFar = textWithNext;
	}
      }
    }
    if (priorWidth != 0) {
      data.add(soFar);
    }

    preferredWidth = maxWidth;
    preferredHeight = fm.getHeight() * data.getCount();

    Control parent = getParent();
    if ((parent != null) && (parent instanceof Scroller)) {
      //System.out.println("scroller invalidate");
      ((Scroller)parent).invalidate();
    }

    repaint();
  }

  // KJT 07/08/01: added setFormattedText for pre-formatted text

  /**
   * Set the text that should be displayed in this text area.
   * The text is presumed already split into lines ending with NL.
   *
   * @param s text to be displayed.
   */
  public void setFormattedText(String s) {
    data = new Vector();
    String line;
    int pos1 = 0, pos2;
    int maxWidth = 0, width;
    FontMetrics fm = getFontMetrics(font);

    while ((pos2 = s.indexOf('\n',pos1)) >= 0) {
      line = s.substring(pos1,pos2);
      width = fm.getTextWidth(line);
      if (width > maxWidth)
        maxWidth = width;
      data.add(line);
      pos1 = pos2 + 1;
    }
    line = s.substring(pos1);
    width = fm.getTextWidth(line);
    if (width > maxWidth)
      maxWidth = width;
    data.add(line);

    preferredWidth = maxWidth;
    preferredHeight = fm.getHeight() * data.getCount();

    Control parent = getParent();
    if ((parent != null) && (parent instanceof Scroller)) {
      ((Scroller)parent).invalidate();
    }

    repaint();
  }

  /**
   * Return the preferred horizontal scroll increment.  By default,
   * the width of a W and m are averaged to return the width for
   * an "average" character.
   */
  public int getHorizontalScrollIncrement() {
    return horizontalScrollInc;
  }

  /**
   * Return the preferred horizontal scroll increment.  By default,
   * the height of the default font is returned.
   */
  public int getVerticalScrollIncrement() {
    return verticalScrollInc;
  }

  /**
   * Implement the onPaint method to display the segmented text
   * @param g graphics to draw with.
   */
  public void onPaint(Graphics g) {

    FontMetrics fm = getFontMetrics(font);

    // KJT 20/08/02: text rectangle needs cleared before rewriting
    g.setBackColor(backColor);
    g.fillRect (0, 0, this.getRect ().width, this.getRect ().height);
    // KJT 20/02/02: changed setColor to setForeColor
    g.setForeColor(0,0,0);
    g.setFont(font);

    String s;
    int itemHeight = fm.getHeight();
    int count = data.getCount();
    // KJT 20/02/02: don't draw text outside box
    int ymax = this.getRect ().height - itemHeight;

    int ypos = -1*vOffset;
    for(int i=0;i<count;i++) {
      if (ypos < 0)
        ypos += itemHeight;
      else if (ypos > ymax)
        break;
      else {
        s = (String)data.get(i);
        g.drawText(s, -1*hOffset, ypos);
        ypos += itemHeight;
      }
    }
  }

  /**
   * Implement the setViewSize method but do nothing with it.
   * @param viewwidth width of textual display.
   * @param viewheight height of textual display.
   */
  public void setViewSize(int viewwidth, int viewheight) {
  }

  /**
   * Scroll the text area vertically to the give offset.
   *
   * @param offset vertical offset value provided by scrollbar.
   */
  public void verticalScroll(int offset) {
    vOffset = offset;
    repaint();
  }

  /**
   * Scroll the text area horizontally to the give offset.
   *
   * @param offset horizontal offset value provided by scrollbar.
   */
  public void horizontalScroll(int offset) {
    hOffset = offset;
    repaint();
  }

  /**
   * Set the preferred size of the text area.  This method allows
   * the programmer to set the preferred size of the text area to
   * override what is computed when the text is segmented in the
   * setText method.
   * @param d dimension that text area should take on.
   */
  public void setPreferredSize(Dimension d) {
    pSize = d;

    Control parent = getParent();
    if ((parent != null) && (parent instanceof Scroller)) {
      //System.out.println("scroller invalidate");
      ((Scroller)parent).invalidate();
    }

  }

  /**
   * Return the preferred size of the text area.  This is either
   * set by the programmer using setPreferredSize() or computed
   * in the setText method.
   * The preferred size is (0,0) until it is computed via setText().
   */
  public Dimension getPreferredSize() {
    if (pSize != null)
      return pSize;
    else
      return new Dimension(preferredWidth,
			   preferredHeight);
  }
}
